import numpy as np
from tornado import gen
from functools import partial
import ast, math, time, logging
from os.path import dirname, join, sys
from concurrent.futures import ThreadPoolExecutor

# Bokeh imports
from bokeh.plotting import figure
from bokeh.server.server import Server
from bokeh.transform import linear_cmap
from bokeh.application import Application
from bokeh.tile_providers import get_provider
from bokeh.layouts import row, column, gridplot
from bokeh.document import without_document_lock
from bokeh.application.handlers.function import FunctionHandler
from bokeh.models import (ColumnDataSource, Button, Select, Slider, CDSView,
                          BooleanFilter, HoverTool, Dropdown, Div, Band,
                          WheelZoomTool, TextInput)

sys.path.insert(0, "compass")
from parameters import FLAGS
from model import CompassModel


class BokehServer():
    """
    Class to run the Bokeh server for visualisation.

    Args:
        doc: a Bokeh Document instance

    Attributes:
        grid_plot: initiates the plot for the grid
        line_plot: initiates one line plot
        update: updates the plots with every time step
        visualise_model: starts the whole visualisation
        grid_values: obtains all the grid values from Mesa.grid
    """

    def __init__(self, doc):
        
        # This is important! Save curdoc() to make sure all threads
        # see the same document.
        self.doc = doc

        # Initialise a CompassModel model with the default setting.
        self.model = CompassModel(**vars(FLAGS), export=False)
        self.params = self.model.params
        self.res_ended = False
        self.school_ended = False
        self.executor = ThreadPoolExecutor(max_workers=2)

        # Initialise layout and populate the plots with the initial
        # configuration
        self.reset_data()
        self.layout()
        
        print('Visualisation running...')

    def agent_filter(self, source, agent_type):
        """
        This function creates a boolean agent filter view for the Bokeh
        visualisation instance.

        Args:
            source (CDS): ColumnDataSource from Bokeh.
            agent_type (str): the agent type to be filtered.

        Returns:
            view: the view that can be used in plots.
        """

        by_type = [True if agent == agent_type else False \
                    for agent in source.data['agent_type']]
        view = CDSView(source=source, filters=[BooleanFilter(by_type)])
        return view

    def customise_grid(self):
        """
        Grid customisation; agent colors, neighbourhood boundaries and schools.

        Todo:
            size should come from a parameter, as they do not scale properly
            with grid size.
        """

        # Agent colours, agent tooltips and grid initialisation
        mapper = linear_cmap(field_name='category',
                             palette=['blue', 'red', 'green', 'orange', 'purple'] ,
                             low=0,
                             high=4)
        TOOLTIPS = [("Residential utility", "@res_utility"),
                    ('Local composition', '@local_comp'),
                    ('Neighbourhood composition', '@n_comp'),
                    ("School utility", "@school_utility"),
                    ('Distance', '@dist_school'),
                    ('School composition', '@s_comp'),
                    ('School composition utility', '@school_comp_utility')]
        hover = HoverTool(names=["households", "schools"], tooltips=TOOLTIPS)
        self.grid = figure(x_range=(self.model.grid.x_min - 1,
                                    self.model.grid.x_max),
                           y_range=(self.model.grid.y_min - 1,
                                    self.model.grid.y_max),
                           tools=[hover, 'tap', 'pan',
                                  WheelZoomTool()],
                           tooltips=TOOLTIPS, output_backend="webgl")

        # Set WheelZoomTool active by default if not lattice
        if self.params['case'].lower() != 'lattice':
            self.grid.toolbar.active_scroll = self.grid.select_one(
                WheelZoomTool)

        # Add a background map using OpenStreetMap (Google Maps is too
        # computationally expensive and cannot zoom properly)
        self.grid.add_tile(get_provider('OSM'))

        self.grid.axis.visible = False
        self.grid.grid.visible = False
        # Function to highlight all households that are currently enrolled in
        # the same school.
        self.source.selected.on_change("indices", self.select_households)

        # Plot households
        self.grid.circle(x="x",
                         y="y",
                         size=5,
                         view=self.household_view,
                         source=self.source,
                         fill_color=mapper,
                         line_color='black',
                         alpha=0.8,
                         nonselection_fill_alpha=0.2,
                         selection_fill_alpha=1,
                         name='households')

        # Plot schools
        self.grid.circle(x="x",
                         y="y",
                         size=7,
                         source=self.source,
                         view=self.school_view,
                         fill_color='yellow',
                         line_color='black',
                         name='schools')

        # Plot neighbourhoods
        self.grid.patches('x',
                          'y',
                          source=self.source,
                          view=self.neighbourhood_view,
                          fill_color=None,
                          line_color="black",
                          line_width=2,
                          hover_alpha=0,
                          hover_line_color=None,
                          name='neighbourhoods',
                          selection_fill_alpha=0)

    def init_grid_plot(self):
        """
        Initiates the grid plot for the visualisation.
        """

        # Create filters to plot households and schools sequentially
        self.household_view = self.agent_filter(self.source, 'household')
        self.school_view = self.agent_filter(self.source, 'school')
        self.neighbourhood_view = self.agent_filter(self.source,
                                                    'neighbourhood')
        self.customise_grid()

    def select_households(self, attr, old, new):
        """
        This function selects all households that are in the same school.

        Args:
            attr: -
            old: -
            new (list): indices that are clicked (can be multiple)
        """
        index = new[0]
        school_id = self.data.iloc[index].school_id
        # List all agents that have the same school id
        same_school = self.data[self.data.school_id == school_id].index
        self.source.selected.indices = list(same_school)

    def dropdown_select(self, event):
        """
        This function creates a dropdown of all schools.

        Args:
            event (str): the selected school
        """

        school_id = int(event.item)
        same_school = self.data[self.data.school_id == school_id].index
        self.source.selected.indices = list(same_school)

    def school_dropdown_func(self, width=100):
        """
        Event listener for the dropdown of all schools.
        """
        schools = self.data[self.data.agent_type == 'school']
        school_locs = [str(num) for num in range(len(schools))]
        self.school_dropdown = Dropdown(label='Select School',
                                        menu=school_locs,
                                        width=width)
        self.school_dropdown.on_click(self.dropdown_select)

    def init_line_plot(self, width=200, height=200, mode='fixed'):
        """
        Initiates the line plot for the server.
        """

        # Create a ColumnDataSource that can be updated at every step.
        TOOLTIPS = [
            ("Residential utility", "@res_utility"),
            ("Residential segregation", "@res_seg"),
            ("School utility", "@school_utility"),
            ("School segregation", "@school_seg"),
        ]
        self.plot = figure(tooltips=TOOLTIPS,
                           y_range=(0, 1),
                           plot_width=width,
                           sizing_mode=mode,
                           title="Neighbourhood/school utility/segregation",
                           output_backend="webgl")

        plot_pars = {
            'Residential utility': {
                'y': 'res_utility',
                'color': 'green',
                'lower': 'res_q5',
                'upper': 'res_q95'
            },
            'Residential segregation': {
                'y': 'res_seg',
                'color': 'blue',
                'lower': None,
                'upper': None
            },
            'School utility': {
                'y': 'school_utility',
                'color': 'orange',
                'lower': 'school_q5',
                'upper': 'school_q95'
            },
            'School segregation': {
                'y': 'school_seg',
                'color': 'purple',
                'lower': None,
                'upper': None
            }
        }

        # Plot lines, markers and bands (for utility only)
        for label in plot_pars.keys():
            y = plot_pars[label]['y']
            color = plot_pars[label]['color']
            lower = plot_pars[label]['lower']
            upper = plot_pars[label]['upper']
            self.plot.line(x='time',
                           y=y,
                           source=self.line_source,
                           line_width=2,
                           color=color,
                           legend_label=label)
            self.plot.circle(x='time',
                             y=y,
                             source=self.line_source,
                             size=5,
                             color=color,
                             legend_label=label)

            if lower is not None:
                band = Band(base='time',
                            lower=lower,
                            upper=upper,
                            source=self.line_source,
                            fill_alpha=0.2,
                            fill_color=color)
            self.plot.add_layout(band)

        self.plot.legend.location = 'top_left'

    def composition_data(self, agent_type='school'):
        """
        Calculates the composition data for schools or neighbourhoods.

        Args:
            agent_type (str): either 'school' or 'neighbourhood'

        Returns:
            DataFrame of fraction blues and reds per agent.
        """

        # Loop over all schools to get the locations and compositions
        cols = ['group0', 'group1', 'dist_school']
        data = self.data[self.data.agent_type == agent_type][cols]

        if agent_type == 'household':
            blues = data[data['group0'] == 1]['dist_school'].values
            reds = data[data['group1'] == 1]['dist_school'].values
            return {'blues': blues, 'reds': reds}
        else:
            totals = data.sum(axis=1).replace(0, 1).values
            fractions = data.div(totals, axis=0)
            blues = fractions['group0'].values
            reds = fractions['group1'].values
            return {'blues': blues, 'reds': reds}

    def init_school_composition_plot(self,
                                     width=200,
                                     height=200,
                                     mode='fixed'):
        """
        Initiates the school composition plot.
        """

        self.school_composition_quads = {}
        self.school_composition_plot = figure(title="School composition",
                                              x_range=(0, 1),
                                              plot_width=width,
                                              sizing_mode=mode,
                                              output_backend="webgl")

        fractions = self.composition_data(agent_type='school')
        for group in fractions.keys():

            hist, edges = np.histogram(fractions[group], density=True, bins=20)
            self.school_composition_quads[
                group] = self.school_composition_plot.quad(
                    top=hist,
                    bottom=0,
                    left=edges[:-1],
                    right=edges[1:],
                    fill_color=group[:-1],
                    line_color="white",
                    alpha=0.7,
                    legend_label=group)

    def init_neighbourhood_composition_plot(self,
                                            width=200,
                                            height=200,
                                            mode='fixed'):
        """
        Initiates the neighbourhood composition plot.
        """
        self.neighbourhood_composition_quads = {}
        self.neighbourhood_composition_plot = figure(
            title="Neighbourhood composition",
            x_range=(0, 1),
            plot_width=width,
            sizing_mode=mode,
            output_backend="webgl")

        fractions = self.composition_data(agent_type='neighbourhood')
        for group in fractions.keys():

            hist, edges = np.histogram(fractions[group], density=True, bins=20)
            self.neighbourhood_composition_quads[group] = \
                self.neighbourhood_composition_plot.quad(
                top=hist, bottom=0, left=edges[:-1], right=edges[1:],
                fill_color=group[:-1], line_color="white", alpha=0.7,
                legend_label=group)

    def init_distribution_plot(self, width=200, height=200, mode='fixed'):
        """
        Initiates the distribution plots for residential and school utility.
        """

        self.distribution_plot = figure(title="Utility distributions",
                                        x_range=(0, 1),
                                        plot_width=width,
                                        sizing_mode=mode,
                                        output_backend="webgl")

        hist_data = self.data[self.data.agent_type == 'household'][[
            'res_utility', 'school_utility'
        ]]

        # Residential utility
        hist, edges = np.histogram(hist_data.res_utility,
                                   density=True,
                                   bins=50)
        self.res_quads = self.distribution_plot.quad(
            top=hist,
            bottom=0,
            left=edges[:-1],
            right=edges[1:],
            fill_color="green",
            line_color="white",
            alpha=0.7,
            legend_label='Residential utility')

        # School utility
        if not self.residential:
            hist, edges = np.histogram(hist_data.school_utility,
                                       density=True,
                                       bins=50)
        # Do not show school utilities yet as they are all zero!
        hist = np.zeros(len(hist))
        self.school_quads = self.distribution_plot.quad(
            top=hist,
            bottom=0,
            left=edges[:-1],
            right=edges[1:],
            fill_color="orange",
            line_color="white",
            alpha=0.7,
            legend_label='School utility')

    def init_distance_plot(self, width=200, height=200, mode='fixed'):
        """
        Initiates the distance plot for the school choice process.
        """
        """
        Initiates the neighbourhood composition plot.
        """
        self.distance_quads = {}
        self.distance_plot = figure(title="Distance utility",
                                    x_range=(0, 1),
                                    plot_width=width,
                                    sizing_mode=mode,
                                    output_backend="webgl")

        fractions = self.composition_data(agent_type='household')
        for group in fractions.keys():

            hist, edges = np.histogram(fractions[group], density=True, bins=20)
            self.distance_quads[group] = \
                self.distance_plot.quad(
                top=hist, bottom=0, left=edges[:-1], right=edges[1:],
                fill_color=group[:-1], line_color="white", alpha=0.7,
                legend_label=group)

    def update_filters(self):
        """
        Updates the view filters for households, schools and neighbourhoods as
        they can change when reset is clicked (i.e., new model instance).
        """

        # Update household filter
        household_filter = [True if agent == 'household' else False for agent \
            in self.source.data['agent_type']]
        self.household_view.filters[0] = BooleanFilter(household_filter)

        # Update neighbourhood filter
        neighbourhood_filter = [True if agent == 'neighbourhood' else False for\
            agent in self.source.data['agent_type']]
        self.neighbourhood_view.filters[0] = BooleanFilter(
            neighbourhood_filter)

        # Update school filter
        school_filter = [True if agent == 'school' else False for agent in \
            self.source.data['agent_type']]
        self.school_view.filters[0] = BooleanFilter(school_filter)

    def reset_data(self):
        """
        Resets the data, could be the initial reset (new sources need to be
        created) or a subsequent one (only update the data).
        """

        # Callback object to run, step and reset the model properly.
        self.residential = True
        self.callback_obj = None
        self.data = self.model.measurements.get_bokeh_vis_data()

        # Check if it's the initial reset (create new sources) or a reset button
        # click (update .data only)
        try:
            self.source.data = self.data
            self.line_source.data = self.data[self.data.agent_type == 'system']
            self.update_filters()
            self.update_data()

        # Reset is clicked --> update .data only
        except AttributeError:
            self.source = ColumnDataSource(self.data)
            self.line_source = ColumnDataSource(
                self.data[self.data.agent_type == 'system'])

    def update_data(self):
        """
        Updates all data sources.
        """

        # Update all plots in the figure
        self.data = self.model.measurements.get_bokeh_vis_data()
        self.source.stream(self.data, len(self.data))
        self.line_source.stream(self.data[self.data.agent_type == 'system'])
        self.school_dropdown_func()

        # Update the utility histograms
        self.update_histograms()

        # Update the composition histograms
        to_update = [self.neighbourhood_composition_quads, 
            self.school_composition_quads, self.distance_quads]

        for quads in to_update:

            # Grab the new data
            if quads == self.neighbourhood_composition_quads:
                hist_data = self.composition_data(agent_type='neighbourhood')
            elif quads == self.school_composition_quads:
                hist_data = self.composition_data(agent_type='school')
            else:
                hist_data = self.composition_data(agent_type='household')

            # Update the bars and edges
            for group in hist_data.keys():

                hist, edges = np.histogram(hist_data[group],
                                           density=True,
                                           bins=20)

                # Update histogram
                quads[group].data_source.data['top'] = hist
                quads[group].data_source.data['left'] = edges[:-1]
                quads[group].data_source.data['right'] = edges[1:]

    def update_histograms(self):

        hist_data = self.data[self.data.agent_type == 'household'][[
            'res_utility', 'school_utility'
        ]]

        hist, edges = np.histogram(hist_data.res_utility,
                                   density=True,
                                   bins=50)
        self.res_quads.data_source.data['top'] = hist
        self.res_quads.data_source.data['left'] = edges[:-1]
        self.res_quads.data_source.data['right'] = edges[1:]

        # Only start to show school utilities when the residential process is
        # finished
        if not self.residential:
            hist, edges = np.histogram(hist_data.school_utility,
                                       density=True,
                                       bins=50)
        else:
            hist = np.zeros(len(hist))

        self.school_quads.data_source.data['top'] = hist
        self.school_quads.data_source.data['left'] = edges[:-1]
        self.school_quads.data_source.data['right'] = edges[1:]

    def blocking_task(self):
        time.sleep(0.5)

    @without_document_lock
    @gen.coroutine
    def unlocked_task(self):
        """
        Needed to make sure that if the reset button is clicked it can go
        inbetween events, otherwise it can be quite slow.
        """
        yield self.executor.submit(self.blocking_task)
        self.doc.add_next_tick_callback(partial(self.step_button))

    def run_button(self):
        """
        Handles the run button clicks, coloring and starts the simulation.
        """
        if self.run.label == 'Run':
            self.run.label = 'Stop'
            self.run.button_type = 'danger'
            self.callback_obj = self.doc.add_periodic_callback(self.unlocked_task, 1000)

        else:
            self.run.label = 'Run'
            self.run.button_type = 'success'
            self.doc.remove_periodic_callback(self.callback_obj)
            
    def step_button(self):
        """
        Checks which process need to be stepped and execute the step. The
        simulate function of the Model instance cannot be used as we need to
        visualise every step.
        """
        if self.run_button=='Stop':
            return

        max_res_steps = self.params['max_res_steps']
        max_school_steps = self.params['max_school_steps']
        school_time = self.model.scheduler.get_time('school')
        residential_time = self.model.scheduler.get_time('residential')

        # If residential is not converged yet or below max steps, do a step
        if (residential_time < max_res_steps and not self.model.res_ended):
            self.model.step(residential=True)
            self.model.res_ended = self.model.convergence_check()
        else:
            self.model.res_ended = True

        # Initial school step needs to be executed
        if (school_time == 0 and self.model.res_ended):
            self.residential = False
            self.model.step(residential=False, initial_schools=True)
            self.model.convergence_check()

        # Normal school steps
        elif (school_time < max_school_steps and self.model.res_ended and not self.model.school_ended):
            self.model.step(residential=False, initial_schools=False)

        if self.model.convergence_check():
            self.model.school_ended = True
            print('School process ended.')

        # Both processes are done/converged
        if (self.model.res_ended and self.model.school_ended):
            self.run_button()
            return

        self.update_data()
        
    def reset_button(self):
        """
        Resets the model and takes the (possible) new parameter values into
        account.
        """

        # Update the parameter values and start a new model
        self.time_series = []
        self.res_ended = False
        self.school_ended = False
        self.params = self.update_pars()
        self.model = CompassModel(**self.params)

        # Stop the model when it is still running while reset is clicked.
        if self.run.label == 'Stop' and self.callback_obj is not None:
            self.doc.remove_periodic_callback(self.callback_obj)
            self.run.label = 'Run'
            self.run.button_type = 'success'

        self.reset_data()
        self.doc.clear()
        self.layout()

    def layout(self):
        """
        Sets up the whole layout; widgets and all plots.
        """

        # Initialise all plots and widgets
        widgets = self.widgets(width=200)

        plot_width = 500
        sizing_mode = 'stretch_height'
        self.init_grid_plot()
        self.init_line_plot(width=plot_width, mode=sizing_mode)
        self.init_distribution_plot(width=plot_width, mode=sizing_mode)
        self.init_school_composition_plot(width=plot_width, mode=sizing_mode)
        self.init_neighbourhood_composition_plot(width=plot_width,
                                                 mode=sizing_mode)
        self.init_distance_plot(width=plot_width, mode=sizing_mode)

        # Row with widgets
        if self.params['case'].lower() == 'lattice':
            width = 420
            split = int(len(widgets) / 2.) + 1
            widget_row = row(
                [column(widgets[:split]),
                 column(widgets[split:])],
                width=width)
        else:
            width = 210
            widget_row = column(widgets, width=width)

        desc = Div(text=open(join(dirname(__file__),
                                  "description.html")).read(),
                   margin=0)
        # Column with all the controls and description
        first_col = column(widget_row, width=width, sizing_mode='fixed')

        # Column with the grid/map
        second_col = column([
            desc,
            row(self.buttons(), sizing_mode='stretch_width'),
            row(self.grid, sizing_mode='stretch_width')
        ],
                            sizing_mode='stretch_width')

        # Column with the plots
        third_col = column([
            self.plot, self.distribution_plot, self.distance_plot,
            self.school_composition_plot, self.neighbourhood_composition_plot
        ])

        vis_layout = gridplot([[first_col, second_col, third_col]],
                              toolbar_location=None)

        self.doc.add_root(vis_layout)
        self.doc.title = "COMPASS"

    def buttons(self, width=100):
        self.run = Button(label="Run", button_type='success', height=32)
        self.run.on_click(self.run_button)
        self.step = Button(label="Step", button_type='primary', height=32)
        self.step.on_click(self.step_button)
        self.reset = Button(label="Reset", button_type='warning', height=32)
        self.reset.on_click(self.reset_button)
        buttons = [self.run, self.step, self.reset]
        return buttons

    def widgets(self, width=100):
        """
        Hardcodes all widgets.
        """

        header_size = '<h3>'

        # Simulation
        self.sleep = Slider(start=0,
                            end=1,
                            value=0.1,
                            step=0.1,
                            title='Time between steps',
                            width=width)
        self.max_move_fraction = Slider(start=0,
                                        end=1,
                                        value=self.params['max_move_fraction'],
                                        step=.05,
                                        title="Fraction moving",
                                        width=width)
        self.max_res_steps = Slider(start=0,
                                    end=1000,
                                    value=self.params['max_res_steps'],
                                    step=10,
                                    title="Residential steps",
                                    width=width)
        self.max_school_steps = Slider(start=0,
                                       end=1000,
                                       value=self.params['max_school_steps'],
                                       step=10,
                                       title="School steps",
                                       width=width)
        self.threshold = Select(title="Convergence threshold",
                                options=['0.001', '0.005', '0.01', '0.02'],
                                value=str(self.params['conv_threshold']),
                                width=width)
        self.stdev = Select(title="Standard deviation of parameters",
                                    options=[str(x / 100.) for x in range(11)],
                                    value=str(self.params['stdev']),
                                    width=width)
        self.temperature = Slider(start=0,
                                  end=500,
                                  step=1,
                                  title="Temperature",
                                  value=self.params['temperature'],
                                  width=width)
        self.window_size = Slider(start=10,
                                  end=50,
                                  value=self.params['window_size'],
                                  step=10,
                                  title="Convergence window size",
                                  width=width)
        self.case = Select(title='Case',
                           options=[
                               'Lattice',
                               'Amsterdam-ethnicity',
                               'Amsterdam-ses',
                               'Amsterdam-income',
                           ],
                           value=str(self.params['case']),
                           width=width)

        if self.params['random_residential'] == 0:
            random_residential = 'False'
        else:
            random_residential = 'True'
        self.random_residential = Select(title="Random residential",
                                         options=['True', 'False'],
                                         value=random_residential,
                                         width=width)

        text = header_size + 'Simulation' + header_size
        simulation_div = Div(text=text, width=width)
        simulation = [
            simulation_div, self.max_move_fraction, self.max_res_steps,
            self.max_school_steps, self.threshold, self.stdev,
            self.window_size, self.temperature, self.case,
            self.random_residential
        ]

        if self.params['case'].lower() != 'lattice':
            text = header_size + 'Parameters' + header_size
            parameter_div = Div(text=text, width=width)
            simulation = [
                parameter_div, self.max_move_fraction, self.max_school_steps,
                self.threshold, self.stdev, self.window_size,
                self.temperature, self.case, self.random_residential
            ]

        # Grid
        self.size = Select(title="Size",
                           options=[str(x * 10) for x in range(1, 16)],
                           value=str(self.params['width']),
                           width=width)
        self.n_neighbourhoods = Select(
            title="Number of neighbourhoods",
            options=[str(x**2) for x in range(1, 16)],
            value=str(self.params['n_neighbourhoods']),
            width=width)
        self.n_schools = Select(title="Number of schools",
                                options=[str(x**2) for x in range(1, 11)],
                                value=str(self.params['n_schools']),
                                width=width)
        self.schools_placement = Select(
            title="School placement",
            options=['evenly_spaced', 'random', 'random_per_neighbourhood'],
            value=str(self.params['schools_placement']),
            width=width)
        self.household_density = Slider(start=0,
                                        end=1,
                                        value=self.params['household_density'],
                                        step=.05,
                                        title="Density",
                                        width=width)
        self.group_dist = Slider(start=0,
                                 end=1,
                                 value=self.params['group_dist'][0][0],
                                 step=.05,
                                 title="Share blue",
                                 width=width)
        self.torus = Select(title="Torus",
                            options=['True', 'False'],
                            value=str(self.params['torus']),
                            width=width)

        if self.params['scheduling'] == 0:
            scheduling = 'replacement'
        else:
            scheduling = 'without_replacement'
        self.scheduling = Select(
            title="Scheduling method",
            options=['replacement', 'without_replacement'],
            value=scheduling,
            width=width)

        text = header_size + 'Environment' + header_size
        environment_div = Div(text=text, width=width)
        grid = [
            environment_div, self.size, self.n_neighbourhoods, self.n_schools,
            self.household_density, self.group_dist, self.scheduling,
            self.torus, self.schools_placement
        ]

        if self.params['case'].lower() != 'lattice':
            grid = [self.scheduling]

        # School
        self.school_capacity = Slider(start=0,
                                      end=3,
                                      value=self.params['school_capacity'],
                                      step=.1,
                                      title="Maximum school capacity",
                                      width=width)
        self.min_capacity = Slider(start=1,
                                      end=200,
                                      value=self.params['min_capacity'],
                                      step=1,
                                      title="Minimum school capacity",
                                      width=width)
        self.school_dropdown_func(width=width)
        text = header_size + 'Schools' + header_size
        school_div = Div(text=text, width=width)
        school = [school_div, self.school_capacity, self.min_capacity, self.school_dropdown]

        if self.params['case'].lower() != 'lattice':
            school = [self.school_capacity, self.min_capacity] #, self.school_dropdown]

        # Household
        self.alpha = TextInput(value=str(self.params['alpha']),
                            title="Alpha",
                            width=width)
        self.utility_at_max = TextInput(value=str(self.params['utility_at_max']),
                            title="Utility at homogeneity",
                            width=width)
        self.optimal_fraction = TextInput(value=str(self.params['optimal_fraction']),
                            title="Optimal fraction",
                            width=width)
        self.p = TextInput(value=str(self.params['p']),
                            title="Location of sigmoid",
                            width=width)
        self.q = TextInput(value=str(self.params['q']),
                            title="Slope of sigmoid",
                            width=width)
        self.radius = Select(title="Radius",
                             options=[str(x) for x in range(1, 11)],
                             value=str(self.params['radius']),
                             width=width)
        self.neighbourhood_mixture = Slider(
            start=0,
            end=1,
            value=self.params['neighbourhood_mixture'],
            step=.05,
            title="Neighbourhood Mixture",
            width=width)
        self.num_considered = Slider(start=1,
                                     end=10,
                                     value=self.params['num_considered'],
                                     step=1,
                                     title="Considered spots",
                                     width=width)
        self.ranking_method = Select(title="Ranking method",
                                     options=['highest', 'proportional'],
                                     value=self.params['ranking_method'],
                                     width=width)

        text = header_size + 'Households' + header_size
        household_div = Div(text=text, width=width)
        household = [
            household_div, self.alpha, self.utility_at_max,
            self.optimal_fraction, self.p, self.q, self.radius, 
            self.neighbourhood_mixture,
            self.num_considered, self.ranking_method, 
        ]

        if self.params['case'].lower() != 'lattice':
            household = [
                self.alpha, self.utility_at_max, self.optimal_fraction,
                self.p, self.q, self.ranking_method
            ]

        widgets = simulation + grid + school + household

        return widgets

    def update_pars(self):
        """
        Updates the parameters.
        """
        new_params = self.params.copy()
        new_params['alpha'] = ast.literal_eval(self.alpha.value)
        new_params['max_move_fraction'] = self.max_move_fraction.value
        new_params['max_school_steps'] = self.max_school_steps.value
        new_params['conv_threshold'] = float(self.threshold.value)
        new_params['utility_at_max'] = ast.literal_eval(self.utility_at_max.value)
        new_params['optimal_fraction'] = ast.literal_eval(self.optimal_fraction.value)
        new_params['school_capacity'] = self.school_capacity.value
        new_params['min_capacity'] = self.min_capacity.value
        new_params['neighbourhood_mixture'] = self.neighbourhood_mixture.value
        new_params['temperature'] = self.temperature.value
        new_params['stdev'] = float(self.stdev.value)
        new_params['window_size'] = int(self.window_size.value)
        new_params['num_considered'] = int(self.num_considered.value)
        new_params['ranking_method'] = self.ranking_method.value
        new_params['p'] = ast.literal_eval(self.p.value)
        new_params['q'] = ast.literal_eval(self.q.value)

        if self.random_residential.value == 'False':
            new_params['random_residential'] = 0
        else:
            new_params['random_residential'] = 1

        if self.scheduling.value == 'replacement':
            new_params['scheduling'] = 0
        else:
            new_params['scheduling'] = 1

        # If coming from another case than lattice, the 
        # n_neighbourhoods and n_schools can be arbitrary nrs.
        new_params['case'] = self.case.value
        if self.case.value.lower()=='lattice':
            
            number = int(self.n_neighbourhoods.value)
            root = math.sqrt(number)
            if int(root + 0.5) ** 2 == number:
                new_params['n_neighbourhoods'] = int(self.n_neighbourhoods.value)
                new_params['n_schools'] = int(self.n_schools.value)
                new_params['max_res_steps'] = self.max_res_steps.value
            else:
                new_params['n_neighbourhoods'] = 25
                new_params['n_schools'] = 25
                new_params['max_res_steps'] = 100

            new_params['width'] = int(self.size.value)
            new_params['height'] = int(self.size.value)
            self.grid.x_range.end = new_params['width']
            self.grid.y_range.end = new_params['height']
            new_params['torus'] = eval(self.torus.value)
            new_params['household_density'] = self.household_density.value
            
            new_params['schools_placement'] = self.schools_placement.value
            new_params['group_dist'] = [[
                self.group_dist.value, 1 - self.group_dist.value
            ]]
        else:
            new_params['max_res_steps'] = 0

        return new_params

    def run_visualisation(self):
        
        apps = {'/': Application(FunctionHandler(BokehServer))}
        server = Server(apps, port=5004)
        # To avoid bokeh's logger spamming
        log = logging.getLogger('bokeh')
        log.setLevel('CRITICAL')
        import warnings
        warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
        server.io_loop.add_callback(server.show, "/")
        server.io_loop.start()




"""
The School class.
"""
import numpy as np
from agents_base import BaseAgent
from typing import Dict, List, ClassVar


class School(BaseAgent):
    """
    The School class.

    Args:
        unique_id (int): unique identifier of the agent.
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all parameter values.

    Attributes:
        unique_id (int): unique identifier of the agent.
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all parameter values.
        capacity (float): the maximum amount of students that can be enrolled.
        total (int): the number of Students at the school
        students (dict[int, 'Student']): all the Student objects enrolled in the school.
        composition (array): the sum of the attribute arrays of all Households
            enrolled in this school.
    """

    _total_schools: ClassVar[int] = 0
    __slots__ = ["idx", "total", "capacity", "min_capacity", "has_space"]

    @classmethod
    def reset(cls) -> None:
        """
        Resets the school counter.
        """
        cls._total_schools = 0

    def __init__(
            self,
            unique_id: int,
            pos: tuple[float, float],
            model: 'CompassModel',
            params: dict
            ):
        super().__init__(unique_id, pos, model, params)

        self.idx: int = School._total_schools
        School._total_schools += 1

        self.total: int = 0
        self.has_space: bool = True
        self.capacity: int = 1 + int(self.params["school_capacity"] * \
                self.params["n_students"] / self.params["n_schools"])
        self.min_capacity = self.params['min_capacity']
        self.students: Dict[int, 'Student'] = {}
        self.composition: np.ndarray = self.new_composition_array()

    def __repr__(self) -> str:
        """
        Returns:
            str: representing the unique identifier of the agent.
        """
        return f"<School object with unique_id: {self.unique_id}>"

    def add_student(self, student: 'Student') -> None:
        """
        Adds a Student object to the list of enrolled students in the School.

        Args:
            student (Student): Student object.
        """
        # Add HOUSEHOLD attributes to the schools' composition
        self.total += 1
        self.composition += student.household.attributes
        self.students[student.idx] = student
        self.has_space = (self.total < self.capacity)

    def remove_student(self, student: 'Student') -> None:
        """
        Removes a Student object from the list of enrolled students in the
        School.

        Args:
            student (Student): Student object.
        """
        # Subtract HOUSEHOLD attributes to the schools' composition
        self.total -= 1
        self.composition -= student.household.attributes  # TODO: zero self.composition?
        self.students.pop(student.idx)
        # after removing a Student, there will always be space
        self.has_space = True

    def get_students(self) -> List['Student']:
        """
        Returns all the students enrolled in the school.

        Returns:
            list: contains all enrolled students.
        """
        return self.students.values()



"""
Implementation of the BaseAgent class.
"""
import numpy as np
from mesa import Agent


class BaseAgent(Agent):
    """
    Extended base class for a MESA model agent.

    Args:
        unique_id (int): unique identifier of the agent.
        pos (tuple): (x,y) coordinates of the agent in the 2D-grid.
        model (CompassModel): CompassModel object.
        params (Argparser): containing all parameter values.
    """

    def __init__(self, unique_id, pos, model, params):

        super().__init__(pos, model)
        self.unique_id = unique_id
        self.pos = pos
        self.model = model
        self.params = params
        self.model.increment_agent_count()

    def __repr__(self):
        """
        Returns:
            str: output string representing the unique identifier of the agent.
        """
        return f"<BaseAgent object with unique_id: {self.unique_id}>"

    def step(self, **kwargs):
        """
        Perform an agent step.

        Note:
            Agents should have this method for the model to work.
        """
        pass

    def advance(self, **kwargs):
        """
        Advance scheduled step.

        Note:
            Agents should have this method for the model to work.
        """
        pass

    def new_composition_array(self):
        """
        Returns:
            array: a new composition array with all values set to zero.
        """
        n_attributes = len(self.params["group_types"][0])
        composition = np.zeros(n_attributes)
        return composition



"""
The school Allocator class.
"""

class Allocator:
    """
    Class that allocates students across schools given their preferences.
    """

    def __init__(self):
        max_moves = 0

    def optimal_school(self, households):
        """
        Allocate student to their optimal school if they are unsatisfied.

        Args:
            households (list): a list of Household objects.

        Note:
            Utility still based on first student only!!!
        """

        # Loop over all students and schools
        for household in households:
            # we need to update the household with information of the first
            # student only. Keep track if we've done that using this variable:
            do_household_update = True
            start_at_school = 0

            for student in household.students:
                current_school = student.school

                # Cannot leave the school if it falls below min_capacity
                if current_school.total <= current_school.min_capacity:
                    break

                for school in student.school_preference[start_at_school:]:

                    # Check if it's the current school
                    if current_school == school:
                        # break means the student stays at its current school
                        break

                    # Check availability and switch
                    if school.has_space:
                        student.new_school(school)
                        # update household
                        if do_household_update:
                            model = household.model

                            # link household and school for the first student
                            # in household_data we need household.students[0].school.unique_id
                            household.school = school
                            household.school_id = school.unique_id
                            household.distance = model.distance_utilities[
                                    household.idx, school.idx
                                    ]
                            do_household_update = False
                        break
                    # As student from the same household have the same preference,
                    # we do not need to check for space in this school anymore
                    start_at_school += 1


    def initial_school(self, households):
        """
        Allocate student to their optimal school.

        Args:
            households (list): a list of Household objects.
        """

        model = households[0].model
        schools_with_space = model.get_agents('schools').copy()

        steps_before_shuffle = len(schools_with_space) / 3
        step = 0

        for household in households:
            # we need to update the household with information of the first
            # student only. Keep track if we've done that using this variable:
            do_household_update = True
            for student in household.students:
                if step == 0:
                    model.random.shuffle(schools_with_space)
                    # reshuffle again when 1/3 of the remaining schools
                    # got a student assigned.
                    # Limit the shuffling if there's less than 10 schools remaining
                    step = max(len(schools_with_space) / 3, 10)
                else:
                    step -= 1

                # assign to the next available school
                school = schools_with_space.pop(0)
                student.new_school(school)

                # update household
                if do_household_update:
                    # link household and school for the first student
                    # in household_data we need household.students[0].school.unique_id
                    household.school = school
                    household.school_id = school.unique_id
                    household.distance = model.distance_utilities[
                            household.idx, school.idx
                            ]
                    do_household_update = False

                if school.has_space:
                    # put back on the list
                    schools_with_space.append(school)



"""
Some utility functions that cannot be in utils.py due
to circular dependencies.
"""

import numpy as np


def calc_comp_utility(
        dest: np.ndarray,
        x: np.ndarray,
        M: np.ndarray,
        f: np.ndarray
        ) -> None:
    """
    Calculates the utility given a normalised composition (0<=x<=1), an
    optimal fraction (0<=f<=1) and utility at homogeneity (0<=M<=1).

    Arguments
        dest  np.ndarray to hold the output
        x
        M
        f
    """
    dest[x <= f] = (x / f)[x <= f]
    dest[x > f] = (M + (1 - x) * (1 - M) / (1 - f))[x > f]



"""
The Model class which initialises the system and all of its components.
"""
import os
import sys
import ijson
import contextlib
import numpy as np
import pandas as pd
from mesa import Model
import geopandas as gpd
from datetime import datetime
from scipy.stats import truncnorm
from scipy.ndimage import convolve
from mesa.space import ContinuousSpace
from shapely.geometry import Point, box
from typing import List, Dict, Generator

sys.path.insert(0, "compass")
from school import School
from utils import Measurements
from household import Household
from neighbourhood import Neighbourhood
from functions import calc_comp_utility
from scheduler import ThreeStagedActivation


@contextlib.contextmanager
def record_time(name: str) -> Generator:
    try:
        start_time = datetime.now()
        yield
    finally:
        print("\n%s: %f" % (name, (datetime.now() - start_time).total_seconds()))


def read_households(path: str) -> pd.DataFrame:
    """
    A fast and efficient parser to read household data from a GeoJSON file.
    Arguments
        path    File to open, passed directly to open()
    Returns
        households (pd.DataFrame) [pos, category, neighbourhood_id]
    """

    # open in binary mode for ijson
    geofile = open(path, "rb")

    # assume floats are good enough,
    # no need for aribrary precision Decimal objects
    geo_households = ijson.items(geofile, "features.item", use_float=True)

    # normal python lists, because appending is fast for them,
    # and we need the columns to create a pandas DataFrame
    coordinates, groups, neighbourhood_ids = [], [], []
    
    for feature in geo_households:
        coordinates.append(feature["geometry"]["coordinates"])
        groups.append(feature["properties"]["group"])
        neighbourhood_ids.append(feature["properties"]["neighbourhood_id"])
    geofile.close()

    return pd.DataFrame(
        {"pos": coordinates, "category": groups, "neighbourhood_id": neighbourhood_ids}
    )


def trunc_normal_sample(means: List, scale: float, size: int, seed=None) -> List[np.ndarray]:
    """
    Samples from a set of truncated normal distributions.
    Arguments:
        means  [n_means] of float
        scale  float
        size   float
    Output:
        sample [n_means] of np.ndarray [size]
    """
    truncnorm._random_state = seed

    sample = [0] * len(means)
    for index, mu in enumerate(means):
        if scale == 0:
            # All samples are equal if the scale is zero.
            sample[index] = np.ones(size) * mu
        else:
            sample[index] = truncnorm.rvs(
                (0 - mu) / scale, (1 - mu) / scale, loc=mu, scale=scale, size=size
            )
    return sample


class CompassModel(Model):
    """
    Model class for school segregation dynamics.
    Args:
        params (Argparser): containing all parameter values.
    Attributes:
        params (dict):
        grid (MultiGrid): MultiGrid object from Mesa.
        scheduler (ThreeStagedActivation): ThreeStagedActivation object.
        agents (list): all agents in the model.
        measurements (Measurements): Measurements object.
        global_composition (dict): containing the total system compositions.
        distance_matrix (list): all the Euclidean distances from one grid cell
            to another.
        nearness_matrix (list): same as above but with normalized distances.
        alpha np.ndarray[n_households]
        temperature float
        utility_at_max np.ndarray[n_households]
        optimal_fraction np.ndarray [n_households]
        neighbourhood_mixture np.ndarray [n_households]
        chosen_indices np.ndarray [n_households]
        school_ended boolean
        res_ended boolean
    Cached attributes:
        kernel (ndarray)
            set: calc_residential_compositions
            used: calc_residential_compositions
        local_compositions (list, ndarray) [n_households, ndarray]
            set: set_agent_parameters, overwritten in load_agents, update_residential
            used: calc_res_utilities
        normalized_compositions (ndarray) [width, height, len(params['group_types'][0])]
            set: calc_residential_compositions
            used: set_agent_parameters
        neighbourhood_compositions (list, ndarray) [n_households]
            set: set_agent_parameters, update_residential
            used: load_agents, calc_res_utilities
        school_compositions (ndarray, float32) [n_households]
            set: set_agent_parameters, calc_school_compositions, update_school
            used: calc_school_utilities
        alpha (ndarray, float32) [n_households]
            set: set_agent_parameters
            used: calc_school_utilities, calc_school_rankings
        temperature (float?)
            set: set_agent_parameters
            used: (calc_school_rankings, residential_ranking but from params)
        utility_at_max (ndarray, float32) [n_households]
            set: set_agent_parameters
            used: calc_res_utilities, calc_school_utilities , calc_school_rankings
            (residential_utility, but from params)
        optimal_fraction (ndarray, float32) [n_households]
            set: set_agent_parameters
            used: calc_res_utilities, calc_school_utilities, calc_school_rankings
        neighbourhood_mixture (ndarray, int) [n_households]  TODO: np.ushort?
            set: set_agent_parameters
            used: calc_res_utilities
        distance_utilities(ndarray) [n_households, n_schools]
            set: set_agent_parameters
            used: calc_school_rankings, update_school
        all_distances:  (ndarray) [n_households, n_schools] FIXME
        household_attrs (ndarray, float) [width, height, len(params['group_types'][0])]
        compositions (ndarray, float32)  as household_attrs
    """

    def __init__(self, export: bool = False, **params: dict):

        super().__init__()

        # Use random from the numpy package for extra functionality
        self.random = np.random.RandomState(self._seed)

        # define various arrays TODO: sort/document those
        self.local_compositions: List = []
        self.neighbourhood_compositions: List = []
        self.alpha: np.ndarray = np.ndarray([])
        self.temperature: float = 0.0
        self.utility_at_max: np.ndarray = np.ndarray([])
        self.optimal_fraction: np.ndarray = np.ndarray([])
        self.neighbourhood_mixture: np.ndarray = np.ndarray([])
        self.school_compositions: np.ndarray = np.ndarray([])
        self.school_ended: bool = False
        self.res_ended: bool = False
        self.distance_utilities: np.ndarray = np.ndarray([])
        self.chosen_indices: np.ndarray = np.ndarray([])

        self.household_attrs: np.ndarray = np.ndarray([])
        self.compositions: np.ndarray = np.ndarray([])
        self.normalized_compositions: np.ndarray = np.ndarray([])

        self.all_distances: np.ndarray = np.ndarray([])
        self.kernel: np.ndarray = np.ndarray([])

        # Initialise the model attributes
        self.set_attributes(params=params, export=export)

        if self.params["case"].lower() != "lattice":
            self.load_agents(self.params["case"].lower())
        else:
            self.create_agents()

        # Get values of the initial configuration
        self.measurements.end_step(residential=True)

        # Calculate global compositions for the segregation calculations
        self.global_composition = self.measurements.neighbourhoods[0, :, 1:].sum(axis=0)

        if self.verbose:
            text = f""" Model initialised:
                NR AGENTS:
                Households: {self.params['n_households']}
                Neighbourhoods: {self.params['n_neighbourhoods']}
                Schools: {self.params['n_schools']}
                In scheduler: {self.scheduler.get_agent_count()}"""
            print(text)

    def set_attributes(self, params: dict, export=False) -> None:
        """
        Sets or calculates all attributes used in the Compass class.
        Args:
            params (Argparser): containing all parameter values.
            export (bool): True if the data needs to be exported or not.
        """

        # Calculate number of households and students
        params["n_households"] = int(
            params["household_density"]
            * (
                params["width"] * params["height"]
                - params["n_neighbourhoods"]
                - params["n_schools"]
            )
        )
        params["n_students"] = int(params["n_households"] * params["student_density"])
        self.params = dict(params)

        # Set tracking attributes
        self.export = export
        # Track segregation over time
        self.segregation: List[float] = []
        self.res_ended = False
        self.school_ended = False
        self.verbose = self.params["verbose"]
        self.agents: Dict[str, List[object]] = {
            "amount": 0,
            "households": [],
            "schools": [],
            "neighbourhoods": [],
        }

        # Initialise other objects
        self.measurements = Measurements(self)
        self.scheduler = ThreeStagedActivation(self)
        self.grid = ContinuousSpace(
            self.params["width"], self.params["height"], torus=self.params["torus"]
        )

    def create_agents(self) -> None:
        """
        Creates the agents when no case study is provided.
        """
        self.neighbourhoods()
        self.schools()
        self.location_to_agent()

        # Compute closest neighbourhoods
        self.closest_neighbourhoods = self.compute_closest_neighbourhoods()

        # Create households
        self.households()

    def check_num_params(self, num_types: int) -> bool:
        """
        Check if lists of parameter values have sufficient
        values for all the group types.

        Todo:
            Generalise in the future
        """ 
        
        # Subset of parameters to check
        pars_to_check = ['optimal_fraction', 'utility_at_max', 
                        'alpha', 'p', 'q']
        for par in pars_to_check:
            if type(self.params[par])!=list:
                return False
            if len(self.params[par][0]) < num_types:
                return False

        return True

    def set_agent_parameters(self, params: dict) -> None:
        """
        Puts the agent parameters in numpy arrays for faster computations.
        Args:
            params (dict): Model parameters which could differ from
                the agent params!
        Todo:
            Parameters should be imported from a config file in the future.
        """
        households = self.get_agents("households")
        n_households = len(households)

        dtype = "float32"
        self.p = np.zeros(n_households, dtype=dtype)
        self.q = np.zeros(n_households, dtype=dtype)
        self.alpha = np.zeros(n_households, dtype=dtype)
        self.temperature = self.params["temperature"]
        self.utility_at_max = np.zeros(n_households, dtype=dtype)
        self.optimal_fraction = np.zeros(n_households, dtype=dtype)
        self.neighbourhood_mixture = np.ones(n_households, dtype=int)

        num_types = len(self.params["group_types"][0])
        self.num_types = num_types # for use in visualisation
        
        # If every parameter of a subset (see check_num_params) has enough
        # values for all types, use these. If not, give them all the first one
        if self.check_num_params(num_types):
            optimal_fractions = np.tile(params["optimal_fraction"][0], (n_households, 1)).T
            alphas = np.tile(params["alpha"][0], (n_households, 1)).T
            utility_at_maxs = np.tile(params["utility_at_max"][0], (n_households, 1)).T
            ps = np.tile(params["p"][0], (n_households, 1)).T
            qs = np.tile(params["q"][0], (n_households, 1)).T

        else:
            optimal_fractions,
            alphas, utility_at_maxs,
            ps, qs = [trunc_normal_sample([params[par][0][0]]*num_types,
                                            scale=params["stdev"],
                                            size=n_households,
                                            seed=self.random
                        ) for par in ['optimal_fraction', 'alpha', 
                        'utility_at_max', 'p', 'q']]

        # Set parameters of every household
        for household in households:
            x, y = household.pos

            # Fill arrays with agent parameter values for faster computations
            self.optimal_fraction[household.idx] = optimal_fractions[
                household.category
            ][household.idx]
            self.alpha[household.idx] = alphas[household.category][household.idx]
            self.utility_at_max[household.idx] = utility_at_maxs[household.category][
                household.idx
            ]
            self.p[household.idx] = ps[household.category][household.idx]
            self.q[household.idx] = qs[household.category][household.idx]

            # Currently only convolution (assumes every household has the same
            # radius) for composition calculations within the lattice case.
            if params["case"].lower() == "lattice":
                self.local_compositions.append(
                    self.normalized_compositions[x, y, household.category]
                )
            else:
                household.params["neighbourhood_mixture"] = 1

            if household.neighbourhood.total > 0:
                norm = 1.0 / household.neighbourhood.total
            else:
                norm = 1.0
            self.neighbourhood_compositions.append(
                household.neighbourhood.composition[household.category] * norm
            )

        self.school_compositions = np.zeros(n_households, dtype=dtype)

        # Distance utilities based on sigmoid function
        if self.params["case"].lower() != "lattice":
            
            self.distance_utilities = 1.0 / (1 + (self.all_distances / self.p[:,np.newaxis]) ** self.q[:,np.newaxis])

            if self.params['subset_schools']:
                N = self.params['num_considered']
                newval = 0
                self.distance_utilities = np.ones(self.all_distances.shape)
                np.put_along_axis(
                    self.distance_utilities, 
                    np.argpartition(self.all_distances, N, axis=1)[:, N:], 
                    newval, axis=1)


    def neighbourhoods(self) -> None:
        """
        Adds the neighbourhood objects to the environment.
        """

        n_neighs = self.params["n_neighbourhoods"]

        # Add neighbourhoods if necessary
        if n_neighs:
            locations = self.choose_locations(n_neighs, "evenly_spaced")
            
            for location in locations:
                size = self.params["width"] / float(n_neighs**0.5 * 2)
                minx, miny = location[0] - size, location[1] - size
                maxx, maxy = location[0] + size, location[1] + size
                shape = box(minx, miny, maxx, maxy)

                # Create the Neighbourhood object, place it on the grid and
                # add it to the scheduler
                neighbourhood = Neighbourhood(
                    self.get_agents("amount"), location, shape, self, self.params
                )
                self.agents["neighbourhoods"].append(neighbourhood)
                self.scheduler.add(neighbourhood)
                self.grid.place_agent(neighbourhood, location)

    def schools(self) -> None:
        """
        Adds the school objects to the environment.
        """

        School.reset()

        # Add schools if necessary
        if self.params["n_schools"]:
            locations = self.choose_locations(
                self.params["n_schools"], self.params["schools_placement"]
            )
            for location in locations:
                # Create the School object, place it on the grid and add it
                # to the scheduler
                school = School(self.get_agents("amount"), location, self, self.params)
                self.agents["schools"].append(school)
                self.scheduler.add(school)
                self.grid.place_agent(school, location)

    def households(self) -> None:
        """
        Adds the household objects to the environment.
        Todo:
            * Place household specific parameters in a attribute called params
        """

        params = self.params
        self.chosen_indices = None  # Only matters for case studies
        self.household_attrs = np.zeros(
            shape=(params["width"], params["height"], len(params["group_types"][0])),
            dtype="float32",
        )

        # Create group types, empty spots and shuffle them both
        n_groups = len(params["group_categories"])
        groups = [
            self.random.choice(
                list(range(0, len(params["group_types"][i]))),
                size=params["n_households"],
                p=params["group_dist"][i],
            )
            for i in range(n_groups)
        ]

        self.grid.empties = set(
            [(x, y) for x in range(params["width"]) for y in range(params["height"])]
        )
        empties = list(self.grid.empties)
        self.random.shuffle(empties)

        # pre-allocate storage for the Housholds
        Household.reset(max_households=params["n_households"])

        for i, position in enumerate(empties[0 : params["n_households"]]):
            household = Household(
                self.get_agents("amount"), position, self, params, groups[0][i]
            )

            # Place households on the grid and add them to the scheduler
            self.agents["households"].append(household)
            self.grid.place_agent(household, position)
            self.grid.empties.discard(position)
            self.scheduler.add(household)
            self.household_attrs[position[0], position[1], :] = household.attributes

        # Calculate AFTER all agents are placed
        self.calc_residential_compositions()
        self.set_agent_parameters(params)
        self.calc_res_utilities()

    def load_agents(self, case="Amsterdam-income") -> None:
        """
        Load the agents from several files.
        Note:
            This function is in progress and works only for 
            the hardcoded cases for now.
        """

        dirname = os.path.dirname(__file__)
        if case.lower() == "amsterdam-ethnicity":
            path = dirname + "/maps/amsterdam-ethnicity"
        elif case.lower() == "amsterdam-ses":
            path = dirname + "/maps/amsterdam-ses"
        elif case.lower() == "amsterdam-income":
            path = dirname + "/maps/amsterdam-income"

        # Load GeoDataFrames
        school_frame = gpd.read_file(path + "/schools.geojson")
        household_frame = read_households(path + "/households.geojson")
        neighbourhood_frame = gpd.read_file(path + "/neighbourhoods.geojson")

        # Create grid
        self.params["torus"] = 0
        self.params["max_res_steps"] = 0 # no residential process
        xmin, ymin, xmax, ymax = neighbourhood_frame.total_bounds
        self.grid = ContinuousSpace(xmax, ymax, self.params["torus"], xmin, ymin)
        self.grid.empties = [(0, 0)] # at least one empty spot for Mesa to work

        # In the file more households could be available to sample from,
        # to create some variation, but only use 100% in a model run
        data = np.load(path + "/distances_perc_of_actual.npz")
        perc_of_actual = data["perc_of_actual"]
        self.all_distances = data["distances"]

        self.scheduler = ThreeStagedActivation(self)

        # More agents are simulated to sample from them and incorporate some
        # randomness in the type and their spatial distribution
        self.params["group_types"][0] = household_frame['category'].unique()
        total_households = len(household_frame)
        actual_households = int(total_households / perc_of_actual)
        self.params["n_households"] = actual_households
        self.params["n_students"] = int(
            self.params["n_households"] * self.params["student_density"]
        )

        # Create agents
        self.create_neighbourhoods(neighbourhood_frame)
        self.create_schools(school_frame)
        self.create_households(household_frame, actual_households)

        if self.verbose:
            print("Setting agent parameters...")
        self.set_agent_parameters(self.params)

        self.local_compositions = self.neighbourhood_compositions
        self.calc_res_utilities()

        if self.params['visualisation']:
            self.school_frame = school_frame
            self.household_frame = household_frame
            self.neighbourhood_frame = neighbourhood_frame

        if self.verbose:
            print("Model loaded!")

    def create_neighbourhoods(self, neighbourhood_frame: pd.DataFrame) -> None:
        """
        Given a GeoDataFrame, this creates all the neighbourhood objects
        """
        if self.verbose:
            print("Creating neighbourhoods...")

        self.params["n_neighbourhoods"] = len(neighbourhood_frame)
        for index, row in neighbourhood_frame.iterrows():
            neighbourhood = Neighbourhood(
                unique_id=index,
                pos=(row.geometry.centroid.xy[0][0], row.geometry.centroid.xy[1][0]),
                shape=row.geometry,
                model=self,
                params=self.params,
            )
            self.agents["neighbourhoods"].append(neighbourhood)
            self.scheduler.add(neighbourhood)
            self.grid.place_agent(neighbourhood, neighbourhood.pos)

    def create_schools(self, school_frame: pd.DataFrame) -> None:
        """
        Given a GeoDataFrame, this creates all the school objects
        """
        School.reset()

        if self.verbose:
            print("Creating schools...")

        self.params["n_schools"] = len(school_frame)
        n_neighbourhoods = self.params["n_neighbourhoods"]
        for index, row in school_frame.iterrows():
            school = School(
                unique_id=index + n_neighbourhoods,
                pos=(row.geometry.xy[0][0], row.geometry.xy[1][0]),
                model=self,
                params=self.params,
            )

            # Maximum capacity can be school dependent in the future
            school.capacity = 1 + int(
                self.params["school_capacity"]
                * self.params["n_students"]
                / self.params["n_schools"]
            )
            self.agents["schools"].append(school)
            self.scheduler.add(school)
            self.grid.place_agent(school, school.pos)

    def create_households(
        self, household_frame: pd.DataFrame, actual_households: int
    ) -> None:
        """
        Given a GeoDataFrame, this creates household objects by randomly
        drawing (without replacement) household metadata rows from the frame.
        Arguments
            actual_households (int)
                The number of households to create.
        Parameters used
            params['random_residential']
        Parameters set
            params['n_households']
        """
        if self.verbose:
            print("Creating households...")

        self.chosen_indices = self.random.choice(
            len(household_frame), size=actual_households, replace=False
        )
        households_sample = household_frame.iloc[self.chosen_indices]

        if self.params["random_residential"]:
            # Randomly shuffle the group of the household
            shuffled = households_sample["category"].values
            self.random.shuffle(shuffled)
            households_sample["category"] = shuffled

        self.params["n_households"] = actual_households
        n_agents = self.params["n_neighbourhoods"] + self.params["n_schools"]
        neighbourhoods = self.get_agents("neighbourhoods")

        self.all_distances = self.all_distances[self.chosen_indices, :]

        # pre-allocate storage for the Housholds
        Household.reset(max_households=actual_households)

        for index, row in enumerate(
            zip(
                households_sample["pos"],
                households_sample["category"],
                households_sample["neighbourhood_id"],
            )
        ):
            household = Household(
                unique_id=index + n_agents,
                pos=row[0],
                model=self,
                params=self.params,
                category=int(row[1]),
                nhood=neighbourhoods[row[2]],
            )
            self.agents["households"].append(household)
            self.scheduler.add(household)
            self.grid.place_agent(household, household.pos)

        self.location_to_agent()

    def location_to_agent(self) -> None:
        """
        Creates a dictionary with the location of the neighbourhoods as key and
        the object itself as value. Schools are not included as they can have
        the same position as a neighbourhood (centroid).
        """
        agents = self.get_agents("neighbourhoods")
        self.location_to_agent = {str(agent.pos): agent for agent in agents}

    def calc_residential_compositions(self) -> None:
        """
        Updates all local residential compositions assuming all households have
        the SAME RADIUS.
        """

        # Determine the kernel of the convolution
        radius = self.params["radius"]
        dim = radius * 2 + 1
        self.kernel = np.ones((dim, dim))
        self.kernel[radius, radius] = 0

        # Should it wrap around the edges or not?
        if self.params["torus"]:
            mode = "wrap"
        else:
            mode = "constant"

        summed = 0
        num_attrs = self.household_attrs.shape[2]
        compositions = np.zeros(shape=self.household_attrs.shape, dtype="float32")

        # Convolution for every household attribute.
        for attr in range(num_attrs):
            compositions[:, :, attr] = convolve(
                self.household_attrs[:, :, attr], self.kernel, mode=mode
            )
            summed += compositions[:, :, attr]
        self.compositions = compositions
        self.normalized_compositions = np.nan_to_num(
            compositions / np.repeat(summed[:, :, np.newaxis], num_attrs, axis=2)
        )

    def calc_school_compositions(self) -> None:
        """
        Calculate the new school compositions for every household and only for
        the first student!
        Note:
            Currently only for the first student!!!
        """

        households = self.get_agents("households")
        for household in households:
            category = household.category
            idx = household.idx
            school = household.school
            if school.total:
                self.school_compositions[idx] = (
                    school.composition[category] / school.total
                )
            else:
                self.school_compositions[idx] = 0.0

    def calc_res_utilities(self) -> None:
        """
        Calculates residential utility at a household its current position and
        given its parameter values.
        """

        b = self.neighbourhood_mixture
        f = self.optimal_fraction
        M = self.utility_at_max
        x = (1 - b) * self.local_compositions + b * self.neighbourhood_compositions

        calc_comp_utility(Household._household_res_utility, x, M, f)

    def calc_school_utilities(self) -> None:
        """
        Calculates school utilities at a student its current school, given
        distance and its other parameter values.
        """

        alpha = self.alpha
        f = self.optimal_fraction
        M = self.utility_at_max
        x = self.school_compositions

        calc_comp_utility(Household._household_school_utility_comp, x, M, f)

        # TODO: This needs to be correct, what distances to use?
        if self.params['subset_schools']:
            Household._household_school_utility = (
                Household._household_school_utility_comp * Household._household_distance
            )
        else:
            Household._household_school_utility = (
                Household._household_school_utility_comp * alpha
            ) + (Household._household_distance * (1 - alpha))

    def calc_school_rankings(
        self, households: List[Household], schools: List[School]
    ) -> None:
        """
        Ranks the schools according to utility.
        Note: this is probably a performance bottleneck; one of the array is
        of size [n_households_moving, n_of_schools].
        Args:
            households (list): list of households the rankings need to be
                calculated for.
            schools (list): list of schools that need to be ranked.
        Todo:
            Schools can differ per household if we only want to look at the
            n-closest schools for example?
        """
        if len(schools) == 0 or len(households) == 0:
            return

        zeros = np.zeros(len(self.params["group_types"][0]))
        compositions = np.array(
            [
                school.composition / school.total if school.total > 0 else zeros
                for school in schools
            ],
            dtype="float32",
        )

        households_data = np.array(
            [(h.idx, h.category) for h in households], dtype=int
        ).T
        households_indices = households_data[0, :]
        households_categories = households_data[1, :]

        # Composition utility calculations
        t = np.take(self.optimal_fraction, households_indices)  # [n_indices]
        M = np.take(self.utility_at_max, households_indices)  # [n_indices]
        x = np.take(
            compositions, households_categories, axis=1
        )  # [n_schools, n_indices]
        composition_utilities = np.where(x <= t, x / t, M + (1 - x) * (1 - M) / (1 - t))

        # Combined (THIS SHOULD BE GENERALISED TO INCLUDE MORE FACTORS)
        utilities = (
            composition_utilities * self.alpha[np.newaxis, households_indices]
            + (
                np.take(self.distance_utilities, households_indices, axis=0)
                * (1 - self.alpha[households_indices, np.newaxis])
            ).T
        )
        
        households_utilities = np.take(
            Household._household_school_utility, households_indices
        )

        method = self.params["ranking_method"].lower()
        if method == "proportional":
            differences = utilities - households_utilities[np.newaxis, :]
            exp_utilities = np.exp(self.temperature * differences)
            transformed = exp_utilities / exp_utilities.sum(axis=0)[np.newaxis, :]

        elif method == "highest":
            transformed = utilities

        else:
            print("Method not implementd.")
            sys.exit()

        # instead of reversing the list, sort the negative values in the list
        ranked_indices = (-transformed).argsort(axis=0, kind='quicksort')

        # necessary to allow indexing with the argsort result
        schools = np.array(schools)

        for i, household in enumerate(households):
            ranking = schools[ranked_indices[:, i]]

            # TODO: can we make school preference a Household property,
            # instead of a Student property? gives ~10% extra speedup
            for student in household.students:
                student.set_school_preference(ranking)

    def get_attributes(self, pos: tuple[float, float]) -> np.ndarray:

        """
        Returns the attribute vector of a given position
        Args:
            pos (tuple): (x,y) coordinates.
        Returns:
            Numpy array: containing all the attributes (all zeros if empty)
        """
        return self.household_attrs[pos[0], pos[1], :]

    def switch_attrs(
        self, pos1: tuple[float, float], pos2: tuple[float, float]
    ) -> None:
        """
        Switches two attribute vectors in the attribute grid by making a copy.
        Args:
            pos1 (tuple): (x,y) coordinates.
            pos2 (tuple): (x,y) coordinates.
        """
        temp = np.copy(self.household_attrs[pos1])
        self.household_attrs[pos1] = self.household_attrs[pos2]
        self.household_attrs[pos2] = temp

    def step(self, residential: bool = False, initial_schools: bool = False) -> None:
        """
        Perform model steps.
        Args:
            residential (bool): True if a residential step needs to be done,
                False (default) means a school step.
            initial_schools (bool): True if an initial school step needs to be
                done, False (default) means a school step.
        """

        # Perform school or residential step.
        self.scheduler.step(residential=residential, initial_schools=initial_schools)

    def simulate(self, res_steps: int = None, school_steps: int = None) -> None:
        """
        Performs #res_steps of residential steps and #school_steps of school
        steps.
        Args:
            res_steps (int): Number of residential steps.
            school_steps (int): Number of school steps.
            initial_schools (bool): True if an initial school step needs to be
                done, False (default) means a school step.
        """

        res_steps = self.params["max_res_steps"]
        while self.scheduler.get_time("residential") < res_steps:

            if self.convergence_check():
                self.res_ended = True
                break
            else: 

                if self.verbose:
                    f = (
                        "Residential process: step "
                        + str(self.scheduler.get_time("residential") + 1)
                        + " from "
                        + str(res_steps)
                    )
                    sys.stdout.write("\r" + f)
                    sys.stdout.flush()

                self.step(residential=True)

        self.res_ended = True
        if self.verbose:
            print()

        school_steps = self.params["max_school_steps"]
        while self.scheduler.get_time("school") < school_steps:

            if self.convergence_check():
                self.school_ended = True
                break
            else: 

                if self.verbose:
                    f = (
                        "School process: step "
                        + str(self.scheduler.get_time("school") + 1)
                        + " from "
                        + str(school_steps)
                    )
                    sys.stdout.write("\r" + f)
                    sys.stdout.flush()

                if self.scheduler.school_steps == 0:
                    self.step(residential=False, initial_schools=True)
                else:
                    self.step(residential=False, initial_schools=False)

        if self.verbose:
            print()
            print("Processes ended")
        self.export_data(self.export)

    def convergence_check(self) -> bool:
        """
        Checks if the processes have converged.
        Returns: True if converged.
        """
        window_size = self.params["window_size"]
        time = self.scheduler.get_time()
        school_time = self.scheduler.get_time("school")

        # Check what type of segregation to calculate (i.e., which of the
        # processes is running)
        if not self.res_ended:
            self.segregation.append(
                self.measurements.calculate_segregation(
                    type="bounded_neighbourhood", index="Theil"
                )
            )
        else:
            self.segregation.append(
                self.measurements.calculate_segregation(type="school", index="Theil")
            )

        # Wait until there is enough steps in the school process
        if (self.res_ended and school_time <= window_size):
            return False

        # Check all metrics in the window size and check if they are below
        # the convergence threshold
        if time >= window_size:
            utilities = self.measurements.households[
                time - window_size + 1 : time + 1, :, 2
            ]
            means = utilities.mean(axis=1)
            stds = utilities.std(axis=1)

            metrics = np.vstack(
                (means, stds, self.segregation[time - window_size + 1 : time + 1])
            )

            metric_means = np.repeat(
                metrics.mean(axis=1)[:, np.newaxis], window_size, axis=1
            )
            mad = np.abs(metrics - metric_means)
            if np.all(mad < self.params["conv_threshold"]):
                # Start over if the residential process has converged
                self.res_ended = True
                return True

        return False

    def choose_locations(
        self, amount: int, method: str = "evenly_spaced"
    ) -> List[tuple[float, float]]:
        """
        Compute a number of locations to place school and neighbourhood objects.
        Args:
            amount (int): the number of agents to place.
            method (str): one of 'evenly_spaced', 'random', 'random_per_neighbourhood'
        Returns:
            list: containing all the locations in tuple (x,y) format.
        """

        if amount == 0:
            return []

        if method == "evenly_spaced":
            per_side = np.sqrt(amount)
            if per_side % 1 != 0:
                print("Unable to place amount of locations using given method")
                sys.exit(1)

            # Compute locations
            per_side = int(per_side)
            xs = np.linspace(0, self.params["width"], per_side * 2 + 1)[1::2]
            ys = np.linspace(0, self.params["height"], per_side * 2 + 1)[1::2]
            return [(x, y) for x in xs for y in ys]

        elif method == "random":
            locations = []
            i = 0
            while i < amount:
                x_coord = self.random.randint(low=0, high=self.params["width"])
                y_coord = self.random.randint(low=0, high=self.params["height"])
                if (x_coord, y_coord) not in locations:
                    locations.append((x_coord, y_coord))
                    i += 1
            return locations

        elif method == "random_per_neighbourhood":
            locations = []
            width, height = self.params["width"], self.params["height"]
            n_schools = self.params["n_schools"]
            n_neighbourhoods = self.params["n_neighbourhoods"]
            per_side = int(np.sqrt(n_neighbourhoods))
            location_width = width / per_side
            location_height = height / per_side

            # Draw a random sample per neighbourhood as long as there are
            # schools to place
            i = 0
            while i < max(n_neighbourhoods, n_schools):
                y_low = 0
                for col in range(per_side):
                    x_low = 0
                    y_high = int((1 + col) * location_height)

                    for row in range(per_side):

                        x_high = int((1 + row) * location_width)

                        if x_high >= width:
                            x_high = width - 1
                        elif y_high >= height:
                            y_high = height - 1

                        x_coord = self.random.randint(low=x_low, high=x_high)
                        y_coord = self.random.randint(low=y_low, high=y_high)

                        # Check if the coordinates haven't already been sampled
                        while (x_coord, y_coord) in locations:
                            x_coord = self.random.randint(low=x_low, high=x_high)
                            y_coord = self.random.randint(low=y_low, high=y_high)

                        locations.append((x_coord, y_coord))
                        x_low = x_high + 1
                        i += 1
                    y_low = y_high + 1

            # Shuffle all locations if n_schools <= n_neighbourhoods, otherwise
            # shuffle only the remainder
            if n_schools <= n_neighbourhoods:
                self.random.shuffle(locations)
            else:
                divider = int(n_schools / n_neighbourhoods)
                remainder = n_schools % n_neighbourhoods
                first_locations = locations[: n_neighbourhoods * divider]
                rest_locations = locations[n_neighbourhoods * divider :]
                self.random.shuffle(rest_locations)
                locations = first_locations + rest_locations[:remainder]

            return locations

        print(f"Unknown method in choose_locations {method} __file__:__line__")
        sys.exit(-1)

    def compute_closest_neighbourhoods(self) -> dict:
        """
        Compute distance from all grid cells to all schools and all
        neighbourhood objects.
        Returns:
            dict: of dicts containing all Euclidean distances
        """

        EPS = 1.2e-6
        closest_neighbourhoods = {}
        neighbourhoods = self.get_agents("neighbourhoods")

        for x in range(self.params["width"]):
            for y in range(self.params["height"]):

                # Loop over all neighbourhoods, calculate distance and save
                # closest
                point = Point(x, y)
                for neighbourhood in neighbourhoods:
                    shape = neighbourhood.shape.buffer(EPS)
                    if shape.contains(point):
                        closest_neighbourhoods[str((x, y))] = str(neighbourhood.pos)
                        break

        return closest_neighbourhoods

    def compute_school_distances(self) -> None:
        """
        Computes school distances.
        """
        self.all_distances = np.zeros(
            (self.params["n_households"], self.params["n_schools"])
        )
        school_frame = gpd.GeoSeries(
            [Point(school.pos) for school in self.get_agents("schools")]
        )
        for household in self.get_agents("households"):
            self.all_distances[household.idx, :] = school_frame.distance(
                Point(household.pos)
            )

    def get_agents(self, agent_type: str) -> List[object]:
        """
        Returns list of agents of given type.
        Args:
            agent_type (str): either 'School', 'Neighbourhood', 'Household' or
            'Student'.
        Returns:
            list: containing all the objects of the specified type.
        """
        return self.agents[agent_type]

    def export_data(self, export=False) -> None:
        """
        Export data for visualization.
        """
        if export:
            self.measurements.export_data()

    def increment_agent_count(self) -> None:
        """
        Increment agent count by one.
        """
        self.agents["amount"] += 1



"""
Contains all the parameter values for one simulation.
"""

import argparse

parser = argparse.ArgumentParser()

# General system parameters
parser.add_argument("--width",
                    type=int,
                    default=50,
                    help="Integer value representing the number of grid cells (lattice case only)")
parser.add_argument("--height",
                    type=int,
                    default=50,
                    help="Integer value representing the number of grid cells (lattice case only)")
parser.add_argument("--torus",
                    action="store_true",
                    default=True,
                    help="Create grid as torus if passed, bounded if not specified (lattice case only)")
parser.add_argument("--household_density",
                    type=float,
                    default=0.9,
                    help="Total fraction of grid cells occupied by households (lattice case only)")
parser.add_argument("--student_density",
                    type=float,
                    default=1,
                    help="Average amount of students per household")
parser.add_argument("--max_move_fraction",
                    type=float,
                    default=0.125,
                    help="Maximum fraction of agents allowed to move during a full step")
parser.add_argument("--max_res_steps",
                    type=int,
                    default=100,
                    help="Maximum amount of steps during residential process (lattice case only)")
parser.add_argument("--max_school_steps",
                    type=int,
                    default=200,
                    help="Maximum amount of steps during school process")
parser.add_argument("--conv_threshold",
                    type=float,
                    default=0.01,
                    help="Convergence threshold for the residential and school processes")
parser.add_argument("--window_size",
                    type=int,
                    default=30,
                    help="How many time steps one looks back for convergence.")
parser.add_argument("--ranking_method",
                    type=str,
                    default='proportional',
                    help="The ranking method of the empty residential spots, \
                    one of 'highest' or 'proportional'")
parser.add_argument("--scheduling",
                    type=str,
                    default=1,
                    help="Does not replace the agent if the value is \
                    1 ('without_replacement'), 0 is with replacement. \
                    Agents are randomly shuffled again \
                    after all have had a chance to move.")
parser.add_argument("--case",
                    type=str,
                    default='Amsterdam-ses',
                    help="To implement case studies. Only lattice and several Amsterdam \
                            cases available for now.")
parser.add_argument("--verbose",
                    type=bool,
                    default=False,
                    help="To enable print statements or not.")
parser.add_argument("--save_last_only",
                    type=bool,
                    default=True,
                    help="Only save last time step to save space on disk.")
parser.add_argument("--filename",
                    type=str,
                    default='data/data',
                    help="Filename to save the data.")
parser.add_argument("--temperature",
                    type=float,
                    default=50,
                    help="Controls the random versus deterministic part in the \
                            behavioural logit rule for all households.")

# Spatial object parameters
parser.add_argument("--n_neighbourhoods",
                    type=int,
                    default=25,
                    help="Number of neighbourhood objects to be placed (lattice case only)")
parser.add_argument("--n_schools",
                    type=int,
                    default=25,
                    help="Amount of schools to be place in grid (lattice case only)")
parser.add_argument("--schools_placement",
                    type=str,
                    default="evenly_spaced",
                    help=
                    "Placement method of schools: random, evenly_spaced, random_per_neighbourhood (lattice case only)")

# Household parameters
parser.add_argument("--group_categories",
                    type=str,
                    nargs="+",
                    default=["attr1"],
                    help="Space seperated list of household groups (e.g. ethnicity, income)")
parser.add_argument("--group_types",
                    type=str,
                    nargs="+",
                    default=[["a", "b"]],
                    help=
                    "Space seperated list of strings containing comma seperated list of types within each group")
parser.add_argument("--group_dist",
                    type=str,
                    nargs="+",
                    default=[[0.5, 0.5]],
                    help=
                    "Space seperated list of strings containing comma seperated list of density of subtype")
parser.add_argument("--utility_at_max",
                    type=str,
                    nargs="+",
                    default=[[0.8, 0.8]],
                    help=
                    "Space seperated list of strings containing comma seperated list of density of subtype")
parser.add_argument("--optimal_fraction",
                    type=str,
                    nargs="+",
                    default=[[0.6, 0.4]],
                    help=
                    "Space seperated list of strings containing comma seperated list of density of subtype")
parser.add_argument("--alpha",
                    type=float,
                    default=[[0.5, 0.3]],
                    help="Weight for composition and distance utility")
parser.add_argument("--p",
                    type=int,
                    default=[[2000, 1000]],
                    help="Controls the location of the 0.5 utility for the distance sigmoid")
parser.add_argument("--q",
                    type=int,
                    default=[[2, 4]],
                    help="Controls the slope of the distance sigmoid")
parser.add_argument("--stdev",
                    type=float,
                    default=0.01,
                    help="Standard deviation of the truncated normal \
                    distribution used in sampling.")
parser.add_argument("--num_considered",
                    type=int,
                    default=8,
                    help="How many empty spots a household considers every \
                        residential move.")
parser.add_argument("--radius",
                    type=int,
                    default=1,
                    help="Radius of circle over which neighbours are taken into account (lattice case only).")
parser.add_argument("--neighbourhood_mixture",
                    type=float,
                    default=0,
                    help=
                    "Ratio bounded neighbourhood to local neighbourhood used in Schelling model (lattice case only). \
                        Zero equals fully local, one fully bounded neighbourhood.")

# School parameters
parser.add_argument("--school_capacity",
                    type=float,
                    default=2,
                    help="Capacity of school as fraction of students in encatchment area")
parser.add_argument("--min_capacity",
                    type=int,
                    default=50,
                    help="Minimum capacity of schools.")

# Parameters for experiments
parser.add_argument("--random_residential",
                    type=int,
                    default=0,
                    help="To generate random residential patterns.")
parser.add_argument("--subset_schools",
                    type=int,
                    default=0,
                    help="To only consider a subset of schools (equal to num_considered).")
parser.add_argument("--visualisation",
                    type=bool,
                    default=False,
                    help="Whether to gather data for visualisation (used in notebooks)")

# Parse and reformat if necessary (multiple group categories)
FLAGS, unparsed = parser.parse_known_args()
if type(FLAGS.group_types[0]) == str:
    FLAGS.group_types = [elem.split(",") for elem in FLAGS.group_types]
    FLAGS.group_dist = [elem.split(",") for elem in FLAGS.group_dist]
    FLAGS.group_dist = [[float(element) for element in sublist]
                        for sublist in FLAGS.group_dist]
    FLAGS.homophilies = [elem.split(",") for elem in FLAGS.homophilies]
    FLAGS.homophilies = [[float(element) for element in sublist]
                         for sublist in FLAGS.homophilies]
    FLAGS.utility_at_max = [elem.split(",") for elem in FLAGS.utility_at_max]
    FLAGS.utility_at_max = [[float(element) for element in sublist]
                            for sublist in FLAGS.utility_at_max]
    FLAGS.optimal_fraction = [
        elem.split(",") for elem in FLAGS.optimal_fraction
    ]
    FLAGS.optimal_fraction = [[float(element) for element in sublist]
                              for sublist in FLAGS.optimal_fraction]
